//
//  AppDelegate.m
//  WindowCaptureDemo
//
//  Created by David Gelhar on 10/23/12.
//  Copyright (c) 2012 __MyCompanyName__. All rights reserved.
//

#import "AppDelegate.h"
#import "DDHotKeyCenter.h"
#import "Carbon/Carbon.h"
#include <ApplicationServices/ApplicationServices.h>

@implementation AppDelegate

@synthesize mirrorWindow,sourceWindowDict,frameDelay,mouseTap;
@synthesize preferencesController;

- (id) init {
    if (self = [super init]) {
        sourceWindowDict = nil;
        mirrorWindow = nil;
        mouseTap = nil;
        frameDelay = 0.02;
    }
    return self;
}



- (void)applicationDidFinishLaunching:(NSNotification *)aNotification
{
    
    self.preferencesController = [[PreferencesController alloc] initWithWindowNibName:@"Preferences"];
    
    [self registerHotKey:self.preferencesController.keyCode withModifierFlags:self.preferencesController.modifierFlags];
    
    // Create event tap for mouse events
    CGEventMask mouseMask = ( CGEventMaskBit(kCGEventLeftMouseDown)
                             |CGEventMaskBit(kCGEventLeftMouseUp)
                             |CGEventMaskBit(kCGEventRightMouseDown)
                             |CGEventMaskBit(kCGEventRightMouseUp)
                             |CGEventMaskBit(kCGEventMouseMoved)
                             |CGEventMaskBit(kCGEventLeftMouseDragged)
                             |CGEventMaskBit(kCGEventRightMouseDragged)
                             |CGEventMaskBit(kCGEventMouseMoved));

 
    mouseTap = CGEventTapCreate(kCGSessionEventTap, 
                            kCGHeadInsertEventTap,kCGEventTapOptionDefault,
                            mouseMask,(CGEventTapCallBack) eventTapFunction, (__bridge void *)self);
    if (!mouseTap) {
        NSLog(@"Can't install mouse hook.");
        return;
    }
    CGEventTapEnable(mouseTap, NO); // Tap starts out disabled
    
    CFRunLoopSourceRef mouseEventSrc = CFMachPortCreateRunLoopSource(kCFAllocatorDefault, mouseTap, 0);

    if ( !mouseEventSrc ) {
        NSLog(@"CFMachPortCreateRunLoopSource failed");
        return;
    }
    
    CFRunLoopRef runLoop = CFRunLoopGetCurrent();
    if ( !runLoop ) {
        NSLog(@"CFRunLoopGetCurrent failed");
        return;
    }
    CFRunLoopAddSource(runLoop,  mouseEventSrc, kCFRunLoopDefaultMode);

}

// Clean up - unregister hot key and 
- (void)applicationWillTerminate:(NSNotification *)aNotification
{
    DDHotKeyCenter * c = [[DDHotKeyCenter alloc] init];
    [c unregisterHotKeysWithTarget:self];
    if (mouseTap != nil) {
        CFRelease(mouseTap);
        mouseTap = nil;
    }
}

// Toggle mirroring
- (void) toggleMirrorWithEvent:(NSEvent *)hkEvent {
    if (mirrorWindow != nil) {
        [self endMirroring];
    } else {
        [self beginMirroring];
    }
    
}
// Begin mirroring the topmost window
- (void) beginMirroring {
    if (mirrorWindow != nil) {  
        [self endMirroring];        // can only window one mirror at a time
    }
    
    // locate topmost window
    sourceWindowDict = [self findTopWindow];
    if (sourceWindowDict == nil) {
        NSBeep();
        NSLog(@"No top window found");
        return;
    }

    // create the mirror 
    mirrorWindow = [MirroredWindow windowFromWindowDict:sourceWindowDict];
    
    // Start the refresh timer
    [self performSelector:@selector(tick) withObject:nil afterDelay:frameDelay];
    
    // (re-)enable the mouse event tap
    if (mouseTap != nil) {
        CGEventTapEnable(mouseTap, YES);
    }

}

// Timer has ticked - refresh the window
- (void) tick {
    if (mirrorWindow && [mirrorWindow refresh]) {
        [self performSelector:@selector(tick) withObject:nil afterDelay:frameDelay];
    } else {
        [self endMirroring];    // if refresh fails, tear down the mirror window
    }
}

// Stop mirroring - get rid of window, disable mouse event tap
- (void) endMirroring {

    if (mouseTap != nil) {
        CGEventTapEnable(mouseTap, NO);
    }
    [mirrorWindow orderOut:nil];    // hide the window
    mirrorWindow = nil;
    sourceWindowDict = nil;
    
}
                    
// Find topmost user window, return dictionary describing it
// Return nil if no open windows
                    
- (NSDictionary *)findTopWindow {
    
    // Get list of windows (from all apps, topmost first)
    NSMutableArray *windows = (__bridge NSMutableArray *)CGWindowListCopyWindowInfo(kCGWindowListOptionOnScreenOnly | kCGWindowListExcludeDesktopElements, kCGNullWindowID);

    // find first window in layer 0
    for (NSDictionary *window in windows) {
        NSNumber *layer = [window objectForKey:@"kCGWindowLayer"];
        if ([layer intValue] == 0) { // in standard layer?
            return window;           // yes - return first window found
        }
    }
    return nil;
}

-(IBAction)showPreferences:(id)sender{
    if(!self.preferencesController) {
        self.preferencesController = [[PreferencesController alloc] initWithWindowNibName:@"Preferences"];
    }
    [self.preferencesController showWindow:self];
}

//
-(void) registerHotKey:(unsigned short) code withModifierFlags:(NSUInteger) flags {
    
    DDHotKeyCenter * c = [[DDHotKeyCenter alloc] init];
    [c unregisterHotKeysWithTarget:self];   //remove any old one
    
    if (![c registerHotKeyWithKeyCode:code 
                        modifierFlags:flags 
                               target:self action:@selector(toggleMirrorWithEvent:) 
                               object:nil]) {
        NSLog(@"Unable to register hotkey");
    }
}
@end

// Event tap for mouse events. 
//
// Modify any mouse event within our mirrored area to use mirror-image coordinates
//
CGEventRef eventTapFunction(CGEventTapProxy proxy, CGEventType type, CGEventRef event, void *refcon) {
    AppDelegate __weak *appDelegate = (__bridge AppDelegate *) refcon;
   // NSLog(@"tap: event type = %d", type);
    
    
    if (type == kCGEventTapDisabledByTimeout) {         // If event is special notification that tap was disabled by timeout
        CGEventTapEnable(appDelegate.mouseTap, YES);    // turn it back on
    } else if (type == kCGEventLeftMouseDown ||         // is it a mouse event?
               type == kCGEventLeftMouseUp ||
               type == kCGEventRightMouseDown ||
               type == kCGEventRightMouseUp ||
               type == kCGEventMouseMoved ||
               type == kCGEventLeftMouseDragged ||
               type == kCGEventRightMouseDragged ||
               type == kCGEventMouseMoved) {
        
        CGPoint location = CGEventGetUnflippedLocation(event);  // note - need to use flipped location for contains test
        CGRect mirrorRect = CGRectZero;
        if (appDelegate.mirrorWindow) {
            mirrorRect = appDelegate.mirrorWindow.frame;
        }
        if (CGRectContainsPoint(mirrorRect, location)) {    // mouse event within mirrored area?
 
            // step 1 - slide left so right edge of image is on x=0
            location.x -= mirrorRect.origin.x + mirrorRect.size.width;
                            
            // step 2 - horizontal flip; now x = [0..mirrorRect.size.width]
            location.x = - (location.x);
            
            // step 3 - slide flipped image back into place
            location.x += mirrorRect.origin.x;
            
            // Correct for use of "GetUnflippedLocation"
            // (window coordinates to have Y=0 at lower left (NS* operations))
            CGFloat screenHeight = [[[NSScreen screens] objectAtIndex:0] frame].size.height;
            location.y = screenHeight - location.y;
            
           // NSLog(@"Flipped location = %f,%f",location.x,location.y);         
            CGEventSetLocation(event, location);    // update mouse location in event
            
        } 
        
    } 
    
    return event;   

}

