#import "NSString+KeyCodeTranslator.h"


@implementation NSString (KeyCodeTranslator)

+ (NSString *)stringForKeyCode:(unsigned short)keyCode withModifierFlags:(NSUInteger)modifierFlags
{
	TISInputSourceRef currentKeyboard = TISCopyCurrentKeyboardInputSource();
	CFDataRef uchr = (CFDataRef)TISGetInputSourceProperty(currentKeyboard, kTISPropertyUnicodeKeyLayoutData);
	const UCKeyboardLayout *keyboardLayout = (const UCKeyboardLayout*)CFDataGetBytePtr(uchr);
	
    // build prefix string from modifiers
    NSString *prefix = @"";    
    if (modifierFlags & NSControlKeyMask) {
	prefix = @"^ ";
    }
    if (modifierFlags & NSAlternateKeyMask) {   // option key
        prefix = [NSString stringWithFormat:@"%@⌥ ", prefix];
    }
    if (modifierFlags & NSCommandKeyMask) {
        prefix = [NSString stringWithFormat:@"%@⌘ ", prefix];
    }
    if (modifierFlags & NSShiftKeyMask) {
        prefix = [NSString stringWithFormat:@"%@⇧ ", prefix];
    }
  

	if(keyboardLayout) {
		UInt32 deadKeyState = 0;
		UniCharCount maxStringLength = 255;
		UniCharCount actualStringLength = 0;
		UniChar unicodeString[maxStringLength];
		
                // note - translate the *unmodified* key
		OSStatus status = UCKeyTranslate(keyboardLayout,
                                                keyCode, kUCKeyActionDown, 0 /*(UInt32)modifierFlags*/,
                                                LMGetKbdType(), 0,
                                                &deadKeyState,
                                                maxStringLength,
                                                &actualStringLength, unicodeString);
		
		if(status != noErr)
			NSLog(@"There was an %s error translating from the '%d' key code to a human readable string: %s", GetMacOSStatusErrorString(status), status, GetMacOSStatusCommentString(status));
		else if(actualStringLength > 0) {	
                    NSString *result = nil;
                
                    // Replace certain characters with user friendly names, e.g. Space, Enter, Tab etc.
                    NSUInteger i = 0;
                    while(i <= NumberOfUnicodeGlyphReplacements && result == nil) {
                            if(mapOfNamesForUnicodeGlyphs[i].glyph == unicodeString[0])
                                    result = NSLocalizedString(([NSString stringWithFormat:@"%s", mapOfNamesForUnicodeGlyphs[i].name, nil]), @"Friendly Key Name");
                            
                            i++;
                    }
                    if (result == nil) { // no substitution found 
                        result = [NSString stringWithCharacters:unicodeString length:(NSInteger)actualStringLength];
                    }

                    return [NSString stringWithFormat:@"%@%@", prefix, result];
		} else
			NSLog(@"Couldn't find a translation for the '%d' key code", keyCode);			
	} else
		NSLog(@"Couldn't find a suitable keyboard layout from which to translate");
	
	return nil;
}

@end